APD_PERCENTAGES = [10,20,30,40,50,60,70,80,90,25,75]

MAX_STIM_AP_DELAY_MS = 200

LOW_PASS_FREQ_HZ = 500
LOW_PASS_ORDER = 4

COLS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'



try:
    from myokit.formats.axon import AbfFile
    useMypokit = True
except:
    useMyokit = False

import numpy as np
import scipy as sp
import scipy.signal
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.ticker import MultipleLocator
from openpyxl import Workbook
from openpyxl.styles import Font #, Color
from openpyxl.worksheet.table import Table, TableStyleInfo
import os
import sys
import time
from math import ceil

def qspline(y1, y2, k1, k2, num):
    a = k1 * num - (y2 - y1)
    b = -k2 * num + (y2 - y1)
    
    t = np.linspace(0, 1, num)

    return (1 - t) * y1 + t * y2 + t * (1-t) * ( (1-t)*a + t*b )

def nleo(sig):
    return np.square(sig) - np.roll(sig, 1) * np.roll(sig, -1)

# TODO define threshold better
def detect_stims(sig, fs, min_stim_dist_ms=10, high_pass_order=4, high_pass_f0_Hz=200, threshold=100):
    fN = fs / 2
    b, a = sp.signal.butter(high_pass_order, high_pass_f0_Hz/fN, 'highpass')
    fsig = sp.signal.filtfilt(b, a, sig)
    nlsig = nleo(fsig)
    
    dist = int(min_stim_dist_ms * fs / 1e3)
    
    min_height = max(100, np.max(nlsig[1:-1]) / 2)
    #print(min_height)
    p, pd = sp.signal.find_peaks(nlsig[1:-1], height=min_height, distance=dist)
    p += 1
    
    ##sfreq = (np.average(np.diff(p)) / fs) if len(p) else 0
    ##print('Detected {} stimuli at {:.3f} Hz'.format(p.shape[0], sfreq))
    
    return p

def detect_aps(sig, fs, min_ap_duration_ms=50, min_ap_amplitude_mV=40, low_pass_order=4, low_pass_f0_Hz=500):
    fN = fs / 2
    b, a = sp.signal.butter(low_pass_order, low_pass_f0_Hz/fN, 'lowpass')
    fsig = sp.signal.filtfilt(b, a, sig)
    nlsig = nleo(fsig)
    
    duration_samples = int(min_ap_duration_ms * fs / 1e3)
    
    peaks, _ = sp.signal.find_peaks(nlsig[1:-1], distance=duration_samples)
    peaks += 1
    
    ph = list()
    
    for p in peaks:
        left = max(0, p-duration_samples)
        right = min(fsig.shape[0], p+duration_samples)
        base = np.median(sig[left:p]) + min_ap_amplitude_mV * 0.1
        if np.count_nonzero(sig[p:right] >= base) > 0.95 * duration_samples:
            if np.max(sig[p:right]) - np.min(sig[left:p]) > min_ap_amplitude_mV:
                ph.append(p)
            
    ph = np.array(ph, dtype=int)
    
    ##sfreq = (np.average(np.diff(ph)) / fs) if len(ph) else 0
    ##print('Detected {} AP at {:.3f} Hz'.format(ph.shape[0], sfreq))
    
    return ph

def remove_stimuli(signal, stimuli, fs, copy=True, max_stim_duration_ms=25):

    fsig = signal.copy() if copy else signal
    
    if len(stimuli) == 0:
        return fsig if copy else signal, []
    
    half_window_samples = int(max_stim_duration_ms * fs / 1e3) // 2
    tenth_window_samples = max(half_window_samples // 5, 10)
    
    nlsig = sp.ndimage.gaussian_filter1d(nleo(fsig), 2)
    slope = np.diff(fsig)

    stims = []
    
    for s in stimuli:
        left = max(0, s - half_window_samples)
        right = min(fsig.shape[0], s + half_window_samples)
        med, std = np.median(nlsig[left:right]), np.std(nlsig[left:right])
        
        pk, pd = sp.signal.find_peaks(nlsig[left:right], height=std, prominence=0)

        if right == fsig.shape[0] and len(pk) < 1:
            start = left
            end = right
        else:
            start = pd['left_bases'][0] + left  # - 1
            #end = pd['right_bases'][-1] + left # + 1
            end = pk[-1] + left + 1
        
        try:
            end += np.where(np.abs(fsig[end:]) < 100)[0][0]
        except:
            pass # TODO: no value below 100 mV - super bad!
        
        try:
            end += np.where(slope[end:right] > 0)[0][0]
        except:
            pass # TODO: no positive slope after stimulus == bad!
        # Also, we should probably consider the voltage at the end of the stimulus...
        
        if end > right:            
            right = min(fsig.shape[0], end + 30)
        
        # TODO: case of start-left ~= 0 # TODO hardcoded 100 samples
        med_slope = np.average(slope[start-100:start]) # np.median(slope[left:start])
        ##std_slope = np.std(slope[left:start])
        ##sqd = np.zeros([right-end])
        ##for i in range(right-end):
        ##    extrapolated = i*med_slope + fsig[start]
        ##    #sqd[i] = np.sqrt((fsig[i+end]-extrapolated)**2)
        ##    sqd[i] = (fsig[i+end]-extrapolated)**2/(1+fsig[start])**2
        ##    sqd[i] += (slope[i+end]-med_slope)**2/(1+med_slope)**2 * 100 # TODO hardcoded weight
        
        # TODO very unelegant and noise-sensitive...
        #for maxrs in range(remslope.shape[0]):
        #    if remslope[maxrs] <= 0:
        #        break
        
        ##wh = (np.cumsum(sqd[::-1]) / np.arange(1, sqd.shape[0]+1))[::-1]
        #end += min(maxrs, np.argmin(sp.signal.medfilt(wh)))
        
        remslope = sp.ndimage.gaussian_filter1d(np.diff(fsig[end:right]), 2)
        wh = (remslope - med_slope)**2
        p, pd = sp.signal.find_peaks(wh, height=0, prominence=0)
        # p contains peaks of greatest divergence. Largest one of them should be the upstroke.
        # -> Need to find attachment left of upstroke
        if p.shape[0]:
            lp = np.argmax(pd['peak_heights'])
            end += np.argmin(wh[:p[lp]])
        elif len(wh):
            end += np.argmin(wh)
            
        try:
            startv = np.median(fsig[start-1:start+2])
        except:
            startv = fsig[start]
        
        # TODO: needs to incorporate: voltage, slope, distance
        #srtd = np.argsort(sp.signal.medfilt(wh))
        #for i in srtd:
        #    try:
        #        endv = np.median(fsig[end+i-1:end+i+2])
        #    except:
        #        endv = fsig[end+i]
        #    if abs(endv-startv) < 1:
        #        end += i
        #        break
        #else:
        #end += np.argmin(sp.signal.medfilt(wh))
        try:
            endv = np.median(fsig[end-1:end+2])
        except:
            endv = fsig[end]
        
        
        k1 = np.average(slope[start-20:start]) # TODO 20 hardcoded...
        k2 = np.average(slope[end:end+20])
        
        if end+1 >= fsig.shape[0]:
            end = fsig.shape[0] - 1
            endv = fsig[start]
            k2 = 0
        
        repl = qspline(startv, endv, k1, k2, end-start+1)
        if np.any(np.isnan(repl)):
            repl = np.linspace(fsig[start], fsig[end], end-start+1)
        fsig[start:end+1] = repl
        
        stims.append([start, end+1])

    return fsig, np.array(stims)

def load_file(fn: str, tscl = -1, vmscl = -1):
    
    if not os.path.exists(fn):
        raise Exception('Error: Cannot open datafile.')

    try:
        aps = list()
        f = AbfFile(fn)
        for trace in f:
            aps.append(np.stack([trace[0].times()*1000, trace[0].values()], -1))
    except:
        try:
            data = np.loadtxt(fn, skiprows=1)

            if tscl == -1:
                # Autoscaling of time column to end up in ms
                dt = np.average(np.diff(data[:,0]))
                if dt <= 0.001: # if this was ms, f_s would be >= 1MHz
                    tscl = 1000.
            data[:,0] *= tscl

            if vmscl == -1:
                # Automatic scaling of V column to end up in mV
                vmm = np.abs(np.median(data[:,1]))
                if vmm <= 0.1: # TODO less than mikrovolts?
                    vmscl = 1000.
                elif 0.1 < vmm <= 1.:
                    vmscl = 100.
                elif 1. < vmm <= 10.:
                    vmscl = 10.
                else:
                    vmscl = 1.
            data[:,1] *= vmscl

            aps = [data]
        except:
            print('Unable to load from file', fn)
            raise Exception('Unable to load from file {}'.format(fn))
            
    return aps

def prepare_excel(fn, apdp, fs):
    wb = Workbook()
    ws = wb.active
    
    title = os.path.splitext(os.path.basename(fn))[0]
    if len(title) > 31:
        title = title[-31:]
    
    ws.title = title

    ws['A1'] = 'Filename:'
    ws['A1'].font = Font(bold=True)
    ws['B1'] = fn

    ws['D1'] = '#Traces:'
    ws['D1'].font = Font(bold=True)

    ws['G1'] = 'Sampling freq. (Hz):'
    ws['G1'].font = Font(bold=True)
    ws['H1'] = fs

    ws['A3'] = ''
    ws['B3'] = 'RMP (mV)'
    ws['C3'] = 'Amplitude (mV)'
    ws['D3'] = 'dVdt_max (filt.) (V/s)'
    ws['E3'] = 'dVdt_max (unf.) (V/s)'

    for i, p in enumerate(apdp):
        ws.cell(row=3, column=i+6, value='APD{:d} (ms)'.format(p))
    ws.cell(row=3, column=6+len(apdp), value='AUC90 (mVs)')
    ws.cell(row=3, column=7+len(apdp), value='SNR (dB)')

    row = ws.row_dimensions[3]
    row.font = Font(bold=True)

    ws['A12'] = '#AP'
    ws['B12'] = 'RMP (mV)'
    ws['C12'] = 'Amplitude (mV)'
    ws['D12'] = 'dVdt_max (filt.) (V/s)'
    ws['E12'] = 'dVdt_max (unf.) (V/s)'
    for i, p in enumerate(apdp):
        ws.cell(row=12, column=i+6, value='APD{:d} (ms)'.format(p))
    ws.cell(row=12, column=6+len(apdp), value='AUC90 (mVs)')
    ws.cell(row=12, column=7+len(apdp), value='SNR (dB)')

    row = ws.row_dimensions[12]
    row.font = Font(bold=True)
    
    return wb, ws

def sortzip(list1, list2):
    i, j = 0, 0
    
    result = []
    
    
    while True:
        if i >= len(list1):
            for k in range(j, len(list2)):
                result.append((list2[k], 1, k))
            break
            
        if j >= len(list2):
            for k in range(i, len(list1)):
                result.append((list1[k], 0, k))
            break
        
        if list1[i] <= list2[j]:
            result.append((list1[i], 0, i))
            i += 1
        else:
            result.append((list2[j], 1, j))
            j += 1

    return result

def low_pass_filter(sig, fs, N=LOW_PASS_ORDER, f0_Hz=LOW_PASS_FREQ_HZ):
    sos = sp.signal.butter(N, f0_Hz/fs*2, 'lowpass', output='sos')
    return sp.signal.sosfiltfilt(sos, sig)

def notch_filter(sig, fs, f0_Hz=50, bw_Hz=0.3, N=2, harmonics:int=0):
    if harmonics > 0:
        f0 = np.arange(f0_Hz, min(fs/2, (harmonics+1) * f0_Hz + f0_Hz/2), f0_Hz) / fs * 2
    elif harmonics == -1:
        f0 = np.arange(f0_Hz, fs/2-bw_Hz, f0_Hz) / fs * 2
    else:
        f0 = [f0_Hz / fs * 2]
        
    bw_half = bw_Hz / fs # == (bw / 2) / fs * 2
    
    sigf = sig
    
    for f in f0[::-1]:  # Better to start with high or low freq ? TODO change bw for higher freq?
        sos = sp.signal.butter(N, [f-bw_half, f+bw_half], 'bandstop', output='sos')
        sigf = sp.signal.sosfiltfilt(sos, sigf)
        
    return sigf

def analyze(fn: str, progressCallback=None, errorCallback=print):
    apdp = APD_PERCENTAGES
    if not 90 in apdp:
        apdp.append(90)
    apdp = np.array(sorted(set(apdp)))
    
    nomarker = plt.Line2D([0],[0], color=(.0,.0,.0,.0))
    
    savename = os.path.splitext(fn)[0]
    traces = load_file(fn)

    fss = []
    for ap in traces:
        fss.append(np.round(1000. / np.average(np.diff(ap[:10000,0])), 3))
    fss=np.array(fss)
    if np.any(fss == fss[0]):
        fs = fss[0]
    else:
        fs = 'varying'
        
    wb, ws = prepare_excel(fn, apdp, fs)
    
    start_t = time.time()
    
    count = 0
    with PdfPages('{}.pdf'.format(savename)) as pdf:
        for j, ap in enumerate(traces):
            fs = round(1000. / np.average(np.diff(ap[:10000,0])), 6)  # time is in ms now... -> 1e3/...

            t = ap[:,0]
            vm = ap[:,1]
            dvdt = np.diff(vm)/np.diff(t)

            si = detect_stims(vm, fs)
            nostim, stims = remove_stimuli(vm, si, fs)

            aps = detect_aps(nostim, fs)

            # TODO: calculate maximum dvdt that is possible at certain f0 (with leeway due to -3db)
            # and make sure that we are not close to it later on...
            #vm_f = notch_filter(low_pass_filter(nostim, fs), fs, harmonics=0)
            # TODO notch only when power spectrum at 50Hz is significantly elevated?
            vm_f = low_pass_filter(nostim, fs)
            
            #if save:
            #    np.savetxt(savename + '_filt.tdms', np.stack([t, vm_f]))
            #    #save_ap(savename + '_filt.tdms', t/1000, vm_f)
            #    return t, vm_f
            
            #rmp_v = low_pass_filter(nostim, fs, f0_Hz=30)
            dvdt_f = np.diff(vm_f) / np.diff(t)  # unit is mV/ms

            fig = plt.figure(figsize=(11.7, 8.3)) # A4 landscape format...
            plt.title('{} - Trace {}'.format(savename, j+1))
            plt.plot(t, vm, color='lightblue')
            plt.plot(t, vm_f, color='tab:blue')
            plt.scatter(t[si], vm_f[si], color='tab:red')
            plt.scatter(t[aps], vm_f[aps], color='tab:orange')
            amp = np.max(vm_f) - np.min(vm_f)
            mn, mx = np.min(vm_f) - 0.1 * amp, np.max(vm_f) + 0.1 * amp
            mn = max(mn, -150) # TODO hardcoded safety numbers
            mx = min(mx, 150)
            plt.ylim([mn, mx])
            fig.legend((nomarker, nomarker),(
                '∅ AP frequency: {} Hz'.format(round(1000/np.average(np.diff(t[aps])), 3)),
                '∅ Stim frequency: {} Hz'.format(round(1000/np.average(np.diff(t[si])), 3))
            ))
            plt.tight_layout()
            pdf.savefig()
            plt.close()
            
            zipd = sortzip(si, aps)
            
            stim_interv_ms = 1000 / (np.average(np.diff(si)) / fs)
            stim_ap_delay = int(min(MAX_STIM_AP_DELAY_MS, stim_interv_ms) * fs / 1000)
            
            skip_next = False
            prev_end = 0
            for i, s in enumerate(zipd):
                failed = False
                warnings = []
                
                if skip_next:
                    skip_next = False
                    continue
                
                idx, tp, oidx = s[0], s[1], s[2]
                
                succ = False
                if tp == 0: # stimulus
                    succ = (i+1<len(zipd) and zipd[i+1][1] == 1 and zipd[i+1][0] - idx < stim_ap_delay)
                    #print('Stimulus at {:.2f}s {}'.format(t[idx]/1000, 'succesfull' if succ else 'unsuccesfull'))
                    if not succ:
                        continue
                    # Succesfull stimulus: analyze following AP
                    skip_next = True
                    s_begin, s_end = stims[oidx][0], stims[oidx][1]
                    idx = zipd[i+1][0]
                    n_ap = zipd[i+1][2]
                    n_stim = oidx
                else:
                    #print('Unstimulated AP at {:.2f}s'.format(t[idx]/1000))
                    s_begin, s_end = None, None
                    n_ap = oidx
                    n_stim = None
                    if i < len(zipd)-1 and zipd[i+1][1] == 0: # next is stimulus
                        s_idx, s_oidx = zipd[i+1][0], zipd[i+1][2]
                        
                        if s_idx - idx < stim_ap_delay:   # TODO ap-stim delay for late-detected stims...
                            s_begin, s_end = stims[s_oidx][0], stims[s_oidx][1]
                            skip_next = True
                        
                
                # Window where the AP should be
                left = 0 if s_end is None else s_end
                left = max(left, prev_end)
                
                nxt = i + (2 if succ else 1)
                if nxt < len(zipd):
                    if zipd[nxt][1] == 0: # Next is stimulus
                        right = stims[zipd[nxt][2]][0] - 1 # use left end of next stim
                        # Can't really trust this though, stim could be in the current AP! 
                        if nxt+1 < len(zipd):
                            right = zipd[nxt+1][0] - 1
                        
                    else: # next is AP
                        right = zipd[nxt][0] - 1 # can only use AP detection index...
                        # TODO this is a problem for SNR below!
                        right -= int(20 * fs / 1000)  # -20 ms ...
                else:
                    right = vm_f.shape[0]
                
                if idx < left and not s_begin is None:
                    left = s_begin
                    warnings.append('Removed stimulus artefact might contain parts of upstroke.')
                
                if idx < left:
                    left = idx - int(50 * fs / 1e3)
                    left = max(0, left)
                    
                if idx > right:
                    # next AP and/or next stim seem to be overlaid
                    continue
                    
                gampl = np.max(vm_f[left:right])-np.min(vm_f[left:right])
                if gampl < 40: # TODO hardcoded amplitude
                    # no AP here (TODO warning)
                    errorCallback('  Skipped ', i, n_ap, ' (no signal)')
                    continue
                elif gampl > 500:
                    errorCallback('  Skipped ', i, n_ap, ' (amplitude high)')
                    # Amplitude too high, probably missed stim detection (TODO warning)
                    continue
                    
                if right-left < 20 * fs / 1e3:
                    # Window too short (TODO warning)
                    errorCallbackcallback('  Skipped ', i, n_ap, ' (window)')
                    continue
                
                try:
                    if not left <= idx < right:
                        errorCallback(left, idx, right, t[left], t[idx], t[right])
                        #continue # TODO failed etc.
                        raise Exception('AP is not in the plausible window...')

                    
                    dvdt_left = left
                    if not s_end is None:
                        dvdt_left = max(dvdt_left, s_end + 1)
                    else:
                        dvdt_left = max(0, idx - int(50 * fs / 1e3)) # TODO Hardcoded 50ms window for upstroke
                    #if vm_f[(left + idx) // 2] < -30:  # Hardcoded -30 mV
                    #    dvdt_left = (left + idx) // 2
                    #dvdt_left = max(left, dvdt_left)
                    
                    dvdt_right = right
                    if nxt < len(zipd) and zipd[nxt][1] == 0: # Next is stimulus
                        dvdt_right = stims[zipd[nxt][2]][0] - 1 # use left end of next stim
                    if dvdt_right <= dvdt_left + int(10*fs/1e3):
                        dvdt_right += int(30*fs/1e3)
                    imax = np.argmax(vm_f[dvdt_left:dvdt_right]) + dvdt_left
                    
                    if vm_f[imax] < -10:  # TODO harcoded min ampl.
                        imax = np.argmax(vm_f[dvdt_left:right-10]) + dvdt_left
                        
                    dvdt_right = imax if imax > dvdt_left else right
                    
                    # Only look for dvdt_max after the next minimum during upstroke (artefact of stimulation)
                    #p, _ = sp.signal.find_peaks(-vm_f[dvdt_left:imax-1], prominence=1, width=ceil(1*fs/1e3))
                    # TODO prominence is empirical value
                    p, _ = sp.signal.find_peaks(-dvdt_f[dvdt_left:int(imax+dvdt_left)//2], prominence=2)
                    if len(p):
                        if dvdt_left + p[0] < dvdt_right:
                            dvdt_left += p[0]
                        #print(n_ap, pd['prominences'], pd['widths'])
                    #if len(p) == 1: # TODO: 1st min is often from stim. but should be more systematic here
                    #    dvdt_left = p[0] + dvdt_left
                    #    # Better maybe: only if there's a minimum with a *maximum* before, start from there.
                    #elif len(p) > 1:
                    #    dvdt_left = p[1] + left
                        
                    dvdt_left = max(dvdt_left, imax - int(50 * fs / 1000))  # 50 ms window for usptroke
                    

                    dvdtmax_i = np.argmax(dvdt_f[dvdt_left:dvdt_right]) + dvdt_left
                    
                    
                    #slope = np.average(dvdt_f[left:i_dvdtmax])
                    #print(slope)
                    # NOTE: not using left here because the RMP could be different due to intrinsic depolarization
                    rmp_l = prev_end
                    rmp_r = dvdtmax_i - int(10 * fs / 1e3) if s_begin is None else s_begin - 1
                    if rmp_r <= rmp_l + 1:
                        #rmp_r = dvdtmax_i
                        rmp_l -= int(50 * fs / 1e3) # TODO harcoded RMP window...
                        rmp_l = max(rmp_l, 0)
                        left = min(left, rmp_l) # Make sure this will be in the plot...
                    
                    #rmp_v = vm_f[rmp_l:rmp_r]
                    # Also, prev end is dangerous for slow repolarization...
                    if rmp_r - rmp_l > 16: # padlen for sosfilt...
                        rmp_v = low_pass_filter(vm_f[rmp_l:rmp_r], fs, 4, 30)
                    else:
                        rmp_v = vm_f[rmp_l:rmp_r]
                    rmp_t = t[rmp_l:rmp_r]
                    rmp_slope = np.median(np.diff(rmp_v) / np.diff(rmp_t))

                    #rmp_v = sp.ndimage.gaussian_filter1d(rmp_v, 5) # TODO XXX make this more robust, amybe low pass 30 Hz?
                    
                    show_rmp_ctx = False
                    if np.abs(rmp_slope) > 0.001:  # mV/ms
                        rmp = np.percentile(rmp_v, 1) # TODO hardcoded %; better find out best slope approximation
                        show_rmp_ctx = True
                    else:
                        rmp = np.median(rmp_v)
                        
                    
                    left = max((dvdtmax_i if s_begin is None else s_begin) - int(50*fs/1e3), left)
                    
                    whrmp = np.where(rmp_v <= rmp)[0]
                    if len(whrmp): # Make sure this region is in the plot...
                        left = min(left, whrmp[-1]+rmp_l-int(20*fs/1e3))
                        left = max(left, 0)

                    # APmax:
                    # - top of filtered?
                    # - median of top 3/5 of unfiltered?

                    # TODO: Right may be inside the next AP, we need to correct it
                    apmax_l = dvdtmax_i
                    apmax_r = right
                    apmax_th = max(rmp+20, vm_f[apmax_l]) # TODO hardcoded min amplitude
                    apmax_ab = np.where(vm_f[apmax_l:apmax_r] < apmax_th)[0]
                    if len(apmax_ab) > 0:
                        for n in apmax_ab:
                            if n > 10:    # TODO hardcoded window for noise...
                                apmax_r = n + apmax_l
                    
                    apmax_i = np.argmax(vm_f[apmax_l:apmax_r]) + apmax_l
                    apmax = vm_f[apmax_i]

                    ampl = apmax-rmp

                    upstroke_l, upstroke_r = dvdtmax_i-(apmax_i-dvdtmax_i)//2, apmax_i
                    if not s_begin is None:
                        upstroke_l = s_begin - int(3*fs/1e3)

                    dvdtmax_unf_med3 = np.median(np.sort(dvdt[dvdt_left:upstroke_r])[-3:])
                    dvdtmax_unf_med5 = np.median(np.sort(dvdt[dvdt_left:upstroke_r])[-5:])
                    dvdtmax_f = np.max(dvdt_f[dvdt_left:upstroke_r]) # TODO does this always correspond to the above defined point?

                    # TODO: since we're working with a filtered signal: maybe all_flwing_under is better than first_under
                    apds = -np.ones([len(apdp)], dtype=int)
                    cur = 0
                    for ix in range(apmax_i, right):
                        if vm_f[ix] <= rmp + (100.-apdp[cur]) / 100. * ampl:
                            apds[cur] = ix
                            cur = cur + 1
                            if cur >= len(apdp):
                                break
                    else:
                        warnings.append('Incomplete repolarization.')
                    apd90_i = apds[apdp==90][0]

                    if apd90_i > 0:
                        rmp_after = np.percentile(vm_f[apd90_i:right], 1)
                        if np.abs(rmp-rmp_after) > 0.1 * ampl:
                            warnings.append('Disgreement between pre-RMP and post-RMP > 10% amplitude.')

                    auc90 = -1
                    if apd90_i > 0:
                        auc90end = apd90_i + 1
                        auc90 = sp.integrate.simps(vm_f[dvdtmax_i:auc90end]-rmp, t[dvdtmax_i:auc90end]) / 1000.

                    # TODO other filter frequency for slow part? (30 Hz seems to work well for RMP...)
                    #snr_start = max(dvdtmax_i, s_end) if not s_end is None else dvdtmax_i
                    #snr_end = right
                    #if nxt < len(zipd):
                    #    if zipd[nxt][1] == 0: # Next is stimulus
                    #        snr_end = max(np.max(apds), 
                    #                      stims[zipd[nxt][2]][0] - 1) # use left end of next stim
                    
                    #psig = np.average((vm_f-rmp)[snr_start:right]**2)
                    #pnoise = np.average((vm-vm_f)[snr_start:right]**2)
                    #snr = 10 * np.log10(psig/pnoise)
                    
                    psig = np.average((rmp_v-rmp)**2)
                    pnoise = np.average((vm[rmp_l:rmp_r]-rmp_v)**2)
                    snr = 10 * np.log10(psig/pnoise)

                    #print(apds)

                    #print(dvdtmax_unf_med3, dvdtmax_unf_med5, dvdtmax_f)
                    #upstroke_vel_50 = np.percentile(dvdt_f[upstroke_l: upstroke_r], 50)
                except Exception as e:
                    failed = True
                    errmsg = str(e)
                    tb = sys.exc_info()[2]
                    if not tb is None:
                        errmsg += ' ({})'.format(tb.tb_lineno)
                    errorCallback(j+1, i+1, count, errmsg, left, right, t[left], t[right])
                
                ### PLOTTING ###
                fig = plt.figure(figsize=(11.7, 8.3)) # outside of try due to finally...
                ax = plt.axes((0.1, 0.1, .65, .85))
                try:
                    plt_l = left #if s_begin is None else max(0, s_begin - 3 * (s_end - s_begin))
                    plt_r = max(right, left+1)
                    
                    if not failed and apd90_i > 0:
                        limit = plt_l + (plt_r - plt_l) // 3 * 2
                        if apd90_i < limit:
                            below = np.where(vm_f[apd90_i:plt_r] < rmp + 0.01 * ampl)[0]
                            if len(below):
                                plt_r = max(limit, below[0]+apd90_i+int(50*fs/1e3))

                    plt_t = t[plt_l:plt_r] # / 1000

                    plt.xlabel('Time (ms)')
                    plt.ylabel('Voltage (mV)')

                    tit = 'AP {}'.format(n_ap + 1)
                    if failed:
                        tit += ' - FAILED'
                    plt.title(tit)
                    plt.plot(plt_t, vm[plt_l:plt_r], color='lightblue', zorder=1)
                    
                    if not failed:
                        plt_rmpv, = plt.plot(t[rmp_l:rmp_r], rmp_v, color='tab:green')
                        plt_rmp, = plt.plot([plt_t[0], plt_t[-1]], [rmp] * 2, '--', color='orange')
                        plt_dvmax, = plt.plot(plt_t, [dvdtmax_f * (x-t[dvdtmax_i]) + vm_f[dvdtmax_i] for x in plt_t], '-.', color='orange', linewidth=1)
                        plt_dvmax_unf, = plt.plot(plt_t, [dvdtmax_unf_med3  * (x-t[dvdtmax_i]) + vm_f[dvdtmax_i] for x in plt_t], '-.', color='red', linewidth=1)
                        #if apd90_i > 0:
                        #    plt.plot(t[apd90_i:plt_r], rmp_v[apd90_i:plt_r], color='tab:green', ls=':')

                    plt.plot(plt_t, vm_f[plt_l:plt_r], color='tab:blue', zorder=3)

                    if not failed:
                        if auc90 != -1:
                            plt.fill_between(t[dvdtmax_i:auc90end], vm_f[dvdtmax_i:auc90end], rmp, zorder=-2, color='lightyellow')
                            plt.plot([t[auc90end-1], t[auc90end-1]], [vm_f[auc90end-1], rmp], color='orange', linewidth=1, zorder=1)
                            plt.plot([t[dvdtmax_i], t[dvdtmax_i]], [vm_f[dvdtmax_i], rmp], color='orange', linewidth=1, zorder=2)
                        plt_auc = plt.Polygon([[0,0]], facecolor='lightyellow', edgecolor='orange', ls='solid', lw=1)

                        plt.scatter(t[apmax_i], vm_f[apmax_i], color='red', marker='x', zorder=2)
                        plt.hlines(vm_f[apmax_i], t[apmax_i]-25, t[apmax_i]+25, linestyle='solid', color='red', zorder=2)
                        plt_apmax, = plt.plot([-1], [-1], color='red', marker='x')
                        plt_dvmaxpt = plt.scatter(t[dvdtmax_i], vm_f[dvdtmax_i], color='orange', zorder=10)

                        plt_apdact = list()
                        colors = ['tab:orange', 'tab:green', 'tab:red', 'tab:purple', 'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan']
                        c = 0
                        for p in apds:
                            if p == -1:
                                plt_apdact.append(nomarker)
                            else:
                                plt_apdact.append(plt.scatter(t[p], vm_f[p], color=colors[c], zorder=10))
                                c = (c + 1) % len(colors)

                        fig.legend((plt_rmp, plt_rmpv, plt_apmax, nomarker, plt_dvmax, plt_dvmax_unf, nomarker, *plt_apdact, nomarker, plt_auc, nomarker, nomarker, plt_rmpv), 
                                   (
                                       'RMP: {:.02f} mV'.format(rmp),
                                       'estimated at',
                                       'Amplitude: {:.02f} mV'.format( ampl),
                                       '',
                                       'dV/dt$_{{max}}$ (filt.): {:.02f} V/s'.format(dvdtmax_f),
                                       'dV/dt$_{{max}}$ (unf.): {:.02f} V/s'.format(dvdtmax_unf_med3),
                                       '',
                                       *('APD{:d}: {}'.format(p, ('N/A' if apds[ix] == -1 else '{:.02f} ms'.format((t[apds[ix]] - t[dvdtmax_i])))) for ix, p in enumerate(apdp)),
                                       '',
                                       'AUC90: {:.02f} mV·s'.format(auc90) if auc90 != -1 else 'AUC90: N/A',
                                       '', 'SNR: {:.02f} dB'.format(snr),
                                       'estimated at'
                                   ),
                                   loc='upper right', markerfirst=False, frameon=False,
                                   bbox_to_anchor=(0.98, 0.95)
                                  )

                    ## TODO: RR_pre/_post, stim-ap-delay

                    plt.xlim([plt_t[0], plt_t[-1]])
                    try:
                        mn, mx = np.min(vm_f[plt_l:plt_r]) - 0.1*ampl, np.max(vm_f[plt_l:plt_r]) + 0.1*ampl
                        plt.ylim([mn, mx])
                    except:
                        plt.ylim([-90, 60])
                    #plt.ylim([rmp-0.1*ampl, apmax+0.1*ampl])


                    ax.xaxis.set_major_locator(MultipleLocator(100))
                    ax.xaxis.set_minor_locator(MultipleLocator(20))
                    ax.yaxis.set_major_locator(MultipleLocator(10))
                    ax.yaxis.set_minor_locator(MultipleLocator(2))


                    # TODO set x/y minor ticks to 10 ms, 5 mV and major to 100ms, 20mV ?
                    plt.grid(True, 'both', color='lightgrey', zorder=0)


                    ### UPSTROKE ZOOM BOX ###
                    if not failed:
                        # Placement for upstroke zoom:
                        if dvdtmax_i > plt_l + (plt_r-plt_l)/2.5:
                            bbox = (0.125, 0.6, .2, .3)
                        else:
                            bbox = (0.525, 0.6, .2, .3)

                        #upstroke_r += int(3*fs/1e3)
                        #upstroke_l = dvdtmax_i - int(5*fs/1e3)
                        upstroke_l += int(3*fs/1e3)
                        upstroke_r = dvdtmax_i + int(5*fs/1e3)

                        ax = plt.axes(bbox)
                        plt.title('Upstroke detail')
                        
                        plt.plot(t[upstroke_l:upstroke_r], vm[upstroke_l:upstroke_r], color='tab:blue', zorder=5)
                        if s_end is not None:
                            plt.plot(t[s_begin:s_end], vm_f[s_begin:s_end], color='lightblue', zorder=6)
                        plt.plot(t[upstroke_l:upstroke_r], vm_f[upstroke_l:upstroke_r], '--', color='tab:blue')

                        plt.scatter(t[apmax_i], vm_f[apmax_i], color='red', marker='x', zorder=10)
                        plt.scatter(t[dvdtmax_i], vm_f[dvdtmax_i], color='orange', zorder=10)

                        up_t = t[upstroke_l:upstroke_r]

                        plt.plot(up_t, [dvdtmax_f * (x-t[dvdtmax_i]) + vm_f[dvdtmax_i] for x in up_t], '-.', color='orange', linewidth=1)
                        plt.plot(up_t, [dvdtmax_unf_med3  * (x-t[dvdtmax_i]) + vm_f[dvdtmax_i] for x in up_t], '-.', color='red', linewidth=1)

                        plt.xlim([t[upstroke_l], t[upstroke_r]])

                        v_min = np.min(vm_f[upstroke_l:upstroke_r])
                        v_max = np.max(vm_f[upstroke_l:upstroke_r])
                        up_amp = v_max-v_min

                        plt.ylim([v_min-.1*up_amp, v_max+.1*up_amp])
                        plt.xticks([])
                        plt.yticks([])
                        
                    ### RMP CONTEXT BOX ###
                    if not failed and show_rmp_ctx:
                        if dvdtmax_i > plt_l + (plt_r-plt_l)/2.5:
                            bbox = (0.125, 0.4, .2, .15)
                        else:
                            bbox = (0.525, 0.4, .2, .15)
                            
                        ax = plt.axes(bbox)
                        plt.title('RMP context')
                        
                        ctx_l = int(max(0, plt_l - 1 * fs)) # 1 second ctx left/right
                        ctx_r = int(min(vm_f.shape[0] - 1, plt_r + 1 * fs))
                        
                        plt.plot(t[ctx_l:ctx_r], vm_f[ctx_l:ctx_r], color='lightblue')
                        plt.plot(plt_t, vm_f[plt_l:plt_r], color='tab:blue')
                        plt.plot(t[rmp_l:rmp_r], rmp_v, color='tab:green')
                        plt.plot([t[ctx_l], t[ctx_r]], [rmp] * 2, '--', color='orange')
                        
                        plt.xlim(t[ctx_l], t[ctx_r])
                        
                        ymin = min(np.min(rmp_v), np.percentile(vm_f[ctx_l:ctx_r], 1))
                        ymax = max(rmp+0.05*ampl, np.max(rmp_v))
                        yamp = ymax-ymin
                        plt.ylim([ymin - 0.05 * yamp, ymax + 0.05 * yamp])
                        plt.xticks([])
                        plt.yticks([])

                    if len(warnings):
                        s = 'WARNING:\n'
                        for w in warnings:
                            s += w + '\n'
                        plt.text(0.775, 0.2, s, wrap=True, transform=fig.transFigure, weight='bold')

                    if failed:
                        s = 'ERROR:\n' + errmsg
                        plt.text(0.775, 0.1, s, wrap=True, transform=fig.transFigure, weight='bold')

                    #plt.show()
                    #break

                    
                    pdf.savefig()
                except Exception as e:
                    #raise(e)
                    try:
                        plt.close()
                    except:
                        pass
                    errmsg = str(e)
                    tb = sys.exc_info()[2]
                    if not tb is None:
                        errmsg += ' ({})'.format(tb.tb_lineno)
                    errorCallback(j+1, i+1, count, errmsg)
                    plt.figure(figsize=(11.7, 8.3))
                    plt.text(0.5, 0.5, 'AP {} - ERROR:\n{}'.format(n_ap + 1, str(e)),
                            horizontalalignment='center', verticalalignment='center')
                    pdf.savefig()
                    #pass # TODO warning, insert empty page...
                finally:
                    plt.close()
                
                ### INSERT INTO XLSX ###
                count += 1
                if not failed:
                    data = [count, rmp, ampl, dvdtmax_f, dvdtmax_unf_med3]
                    data.extend((t[apds[ix]] - t[dvdtmax_i]) if apds[ix] != -1 else 'N/A' for ix in range(len(apdp)))
                    data.extend([auc90 if auc90 > -1 else 'N/A', snr])
                else:
                    data = [count, errmsg]
                if len(warnings):
                    s = 'WARNING:'
                    for w in warnings:
                        s += ' ' + w
                    data.append(s)
                ws.append(data)
                
                if failed:
                    prev_end = left  #, idx + 1) could be problematic if the next stim is following immediately...
                    # maybe check above that left is idx-X ms?
                # TODO this could still be problematic...
                elif n_ap + 1 < len(aps):
                    if apd90_i and apd90_i > 0:
                        nxti = min(right, apd90_i+(apd90_i-dvdtmax_i))
                        if apd90_i <= nxti:
                            prev_end = apd90_i
                        else:
                            prev_end = np.argmin(vm_f[apd90_i:nxti]) + apd90_i
                    elif apds[0] != -1:
                        prev_end = apds[np.where(apds > 0)[0][-1]]
                    else:
                        prev_end = (right+left)//2

                if progressCallback is None:
                    print('{:.1f}%, {:.1f}% ({:.1f} APs/sec)'.format((j+1)/len(traces)*100, (i+1)/len(zipd)*100, 
                                                              count/(time.time()-start_t)), end='\r', flush=True)
                else:
                    progressCallback(j, len(traces), i, len(zipd))
        
                    
    ws['E1'] = count

    
    # TODO: is 6 right with ACU and SNR?
    ws['A4'] = 'N'
    ws['A4'].font = Font(bold=True)
    for i in range(6+len(apdp)):
        ws.cell(row=4, column=i+2, value='=COUNT({0}13:{0}{1})'.format(COLS[i+1], 12+count))

    ws['A5'] = 'Min'
    ws['A5'].font = Font(bold=True)
    for i in range(6+len(apdp)):
        ws.cell(row=5, column=i+2, value='=MIN({0}13:{0}{1})'.format(COLS[i+1], 12+count))

    ws['A6'] = 'Max'
    ws['A6'].font = Font(bold=True)
    for i in range(6+len(apdp)):
        ws.cell(row=6, column=i+2, value='=MAX({0}13:{0}{1})'.format(COLS[i+1], 12+count))

    ws['A7'] = 'Median'
    ws['A7'].font = Font(bold=True)
    for i in range(6+len(apdp)):
        ws.cell(row=7, column=i+2, value='=MEDIAN({0}13:{0}{1})'.format(COLS[i+1], 12+count))

    ws['A8'] = 'Average'
    ws['A8'].font = Font(bold=True)
    for i in range(6+len(apdp)):
        ws.cell(row=8, column=i+2, value='=AVERAGE({0}13:{0}{1})'.format(COLS[i+1], 12+count))

    ws['A9'] = 'SD'
    ws['A9'].font = Font(bold=True)
    for i in range(6+len(apdp)):
        ws.cell(row=9, column=i+2, value='=STDEV({0}13:{0}{1})'.format(COLS[i+1], 12+count))


    ws['A10'] = 'SEM'
    ws['A10'].font = Font(bold=True)
    for i in range(6+len(apdp)):
        ws.cell(row=10, column=i+2, value='=STDEV({0}13:{0}{1})/SQRT(COUNT({0}13:{0}{1}))'.format(COLS[i+1], 12+count))

    tab = Table(displayName="Table1", ref="A12:{}{}".format(COLS[len(apdp)+6], 12+count))

    # Add a default style with striped rows and banded columns
    style = TableStyleInfo(name="TableStyleMedium9", showFirstColumn=True,
                           showLastColumn=False, showRowStripes=True, showColumnStripes=True)
    tab.tableStyleInfo = style
    ws.add_table(tab)

    wb.save('{}.xlsx'.format(savename))


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('Error: No datafile.')
        os.system('pause')
        sys.exit(-1)

    fn = sys.argv[1]
    if not os.path.exists(fn):
        print('Error: Can not open datafile.')
        os.system('pause')
        sys.exit(-2)

    try:
        analyze(fn)
    except Exception as e:
        print('Error: ' + str(e))
        os.system('pause')
        sys.exit(-3)
